from os import listdir
from os.path import isfile, join
from typing import List, Tuple

import pandas as pd

from sqltest.datasource.reader import DatasetReader


class CsvDatasetReader(DatasetReader):

    def __init__(self, data_path: str, source_key="source", target_key="target"):
        self._data_path = data_path
        self._source_key = source_key
        self._target_key = target_key

    def read_source_data(self) -> List[Tuple[str, pd.DataFrame]]:
        return self._read_csv(join(self._data_path, self._source_key))

    def read_target_data(self) -> List[Tuple[str, pd.DataFrame]]:
        return self._read_csv(join(self._data_path, self._target_key))

    @staticmethod
    def _read_csv(path: str) -> List[Tuple[str, pd.DataFrame]]:
        csv_files = [f for f in listdir(path) if isfile(join(path, f)) and f.endswith('.csv')]
        for f in csv_files:
            yield f.split('.csv')[0], pd.read_csv(join(path, f))



from abc import ABC
import re
from typing import List

from sqltest.parser.exceptions import VariableNotFound
from sqltest.parser.operation import NothingOperation, Operation, RunnableOperation


class ExecutionEnvironment(object):
    def __init__(self, env: dict):
        self._env = env

    def check_variable(self, var: str) -> str:
        if var in self._env:
            return var
        else:
            raise VariableNotFound(var)

    def get_variable_val(self, var) -> str:
        return self._env.get(self.check_variable(var))

    def set_variable(self, key: str, val: str):
        self._env[key] = val


class ParseStrategy(object):
    def match(self, statement: str) -> bool:
        raise NotImplementedError("this method is not implemented.")

    def convert(self, statement: str, context: ExecutionEnvironment) -> Operation:
        raise NotImplementedError("this method is not implemented.")

    def get_hints(self) -> List[str]:
        raise NotImplementedError("this method is not implemented.")


class AbstractRegexParseStrategy(ParseStrategy, ABC):
    def __init__(self, pattern):
        self.pattern = pattern

    def match(self, statement: str) -> bool:
        return bool(re.search(self.pattern, statement.strip()))

    @staticmethod
    def overwrite_variables(statement: str, context: ExecutionEnvironment) -> str:
        for var in re.findall("\${(\w*)}", statement):
            val = context.get_variable_val(var)
            statement = statement.replace(f'${{{var}}}', val)
        return statement


class VerifyVariableStrategy(AbstractRegexParseStrategy):
    def __init__(self):
        super().__init__("--variable")

    def convert(self, statement: str, context: ExecutionEnvironment) -> Operation:
        context.check_variable(statement.strip().split("variable=")[1])
        return NothingOperation()

    def get_hints(self) -> List[str]:
        return ["--variable=xxx"]


class ExtractConstantStrategy(AbstractRegexParseStrategy):
    def __init__(self):
        super().__init__("--const")

    def convert(self, statement: str, context: ExecutionEnvironment) -> Operation:
        statement = self.overwrite_variables(statement, context)

        statement = statement.split(self.pattern)[1].strip()
        const_key = statement.split("=")[0]
        const_val = statement.split("=")[1]
        context.set_variable(const_key, const_val)
        return NothingOperation()

    def get_hints(self) -> List[str]:
        return ["--const var=val", "--const var=val_${other_var}"]


class ExtractSQLStrategy(AbstractRegexParseStrategy):
    def __init__(self):
        super().__init__("[CREATE|INSERT|SELECT|DROP|SET]")

    def convert(self, statement: str, context: ExecutionEnvironment) -> Operation:
        statement = self.overwrite_variables(statement, context)
        return RunnableOperation(statement.split(";")[0])

    def get_hints(self) -> List[str]:
        return ["CREATE TABLE IF NOT EXISTS xxx"]



from pyspark.sql import SparkSession

from sqltest.datasource.reader import DatasetReader
from sqltest.engine import SparkEngine


def spark_engine(spark: SparkSession, sql_path: str, reader: DatasetReader = None, env: dict = None):
    def wrapper(func):
        def decorate(*args):
            engine = SparkEngine(spark, env)
            dataset_reader = retrieve_reader(args)
            if not dataset_reader:
                raise Exception("DatasetReader is not defined.")

            engine.run(dataset_reader, sql_path)

            func(*args, engine)

        def retrieve_reader(args):
            if not reader:
                for arg in args:
                    if isinstance(arg, DatasetReader):
                        return arg
            else:
                return reader

        return decorate

    return wrapper



from pandas import DataFrame, testing


def assert_frame_equal(left: DataFrame, right: DataFrame, sort_keys=None):
    if sort_keys is None:
        sort_keys = []
    if len(sort_keys):
        left = left.sort_values(by=sort_keys).reset_index(drop=True)
        right = right.sort_values(by=sort_keys).reset_index(drop=True)

    testing.assert_frame_equal(left, right, check_dtype=False, check_index_type=False)



import logging
import re
from typing import List, Tuple

import pandas as pd
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import *

from sqltest.engine.engine import SqlEngine
from sqltest.tools.dataframe_comparator import assert_frame_equal

LOG = logging.getLogger(__name__)


def pandas_to_spark(pandas_df: pd.DataFrame, spark: SparkSession) -> DataFrame:
    columns = list(pandas_df.columns)
    types = list(pandas_df.dtypes)
    struct_list = []
    for column, typo in zip(columns, types):
        struct_list.append(define_structure(column, typo))
    p_schema = StructType(struct_list)
    return spark.createDataFrame(pandas_df, p_schema)


def equivalent_type(f):
    if f == 'datetime64[ns]':
        return TimestampType()
    elif f == 'int64':
        return LongType()
    elif f == 'int32':
        return IntegerType()
    elif f == 'float64':
        return FloatType()
    else:
        return StringType()


def define_structure(string, format_type):
    try:
        typo = equivalent_type(format_type)
    except:
        typo = StringType()
    return StructField(string, typo)


class SparkEngine(SqlEngine):
    def __init__(self, spark: SparkSession, env_variables: dict):
        super().__init__(env_variables)
        self._spark = spark

    def verify_target_dataset(self):
        for (table, dataset) in self.target_dataset:
            result_df = self._query_table(table)
            assert_frame_equal(result_df, dataset, sort_keys=dataset.columns.values.tolist())

    def _execute(self, statements: List[str]):
        self._register_temp_view(self.source_dataset)
        self._execute_sql(statements)

    def _execute_sql(self, statements: List[str]):
        for stat in statements:
            self._create_database_if_not_exist(stat)
            stat = self._inline_data_source_table(stat)
            self._spark.sql(stat)

    def get_target_tables(self) -> List[Tuple[str, pd.DataFrame]]:
        for (table, dataset) in self.target_dataset:
            yield table, self._query_table(table)

    def _query_table(self, table):
        return self._spark.sql(f"SELECT * FROM {table}").toPandas()

    def _register_temp_view(self, datasets: List[Tuple[str, pd.DataFrame]]):
        for (table, dataset) in datasets:
            pandas_to_spark(dataset, self._spark).createOrReplaceTempView(table.replace(".", "_"))
            LOG.info(f'Register temporary view {table.replace(".", "_")}')

    def _verify_target_dataset(self, target_dataset):
        for (table, dataset) in target_dataset:
            result_df = self._query_table(table)
            assert_frame_equal(result_df, dataset, sort_keys=dataset.columns.values.tolist())

    def _inline_data_source_table(self, statement):
        if bool(re.match('(^INSERT|^SELECT)', statement.strip(), re.I)):
            raw_full_table_names = self._extract_source_tables(statement)
            for full_table_name in raw_full_table_names:
                new_table_name = full_table_name.replace(".", "_")
                statement = statement.replace(full_table_name, new_table_name)
                LOG.info(f"Replace table full name from {full_table_name} to {new_table_name}")
            return statement
        else:
            return statement

    @staticmethod
    def _extract_source_tables(statement):
        for table in re.findall("FROM\s+(.*?)\)?\s+", statement, flags=re.I):
            yield table

        for table in re.findall('JOIN\s+(.*?)\s+.*\s+ON', statement, flags=re.I):
            yield table

    def _create_database_if_not_exist(self, statement):
        for db in re.findall("^CREATE\s+[TABLE|IF NOT EXISTS]+\s+(.*?)\.", statement.strip(), flags=re.I):
            self._spark.sql(f"CREATE DATABASE IF NOT EXISTS {db}")
            LOG.info(f"Try to create database: {db}")

# if __name__ == '__main__':
#     print(re.findall("FROM\s+(.*?)\)?\s+", "FROM dw_dwd_dev.subject_scores) ", flags=re.I))
#     print(re.findall("FROM\s+(.*?)\)?\s+", "FROM dw_dwd_dev.subject_scores  test", flags=re.I))
#     print(re.findall("FROM\s+(.*?)\)?\s+", "FROM dw_dwd_dev.subject_scores test", flags=re.I))
#     print(re.findall("FROM\s+(.*?)\)?\s+", "FROM dw_dwd_dev.subject_scores test", flags=re.I))
#     print(re.findall("FROM\s+(.*?)\)?\s*", "FROM dw_dwd_dev.subject_scores", flags=re.I))
#     print(re.findall("FROM\s+(.*?)\)?\s+", "FROM dw_dwd_dev.subject_scores ", flags=re.I))



from typing import List

from sqltest.parser.strategy import ExecutionEnvironment, ParseStrategy

COMMENT_PREFIX = '--'
MASK = "--.*$"''
BEGINNING_MASK = "^(\\s)*--.*$"


def is_comment_statement(line: str):
    return line.lstrip().startswith(COMMENT_PREFIX)


def is_end_of_statement(line: str):
    return line.replace(MASK, "").rstrip().endswith(";")


def normalize(buffer: List[str]):
    return "\n".join([line.replace(BEGINNING_MASK, "") for line in buffer])


class SqlFileReader(object):

    @staticmethod
    def _read_sql_statements(path: str) -> List[str]:
        with open(path, 'r', encoding='utf8') as file:
            lines = file.readlines()
            buffer = []
            for line in lines:
                if is_comment_statement(line):
                    yield line
                elif is_end_of_statement(line):
                    buffer.append(line)
                    yield normalize(buffer)
                    buffer.clear()
                else:
                    buffer.append(line)
            if len(buffer):
                yield normalize(buffer)

    def compile_sql_file(self, path: str, env: ExecutionEnvironment, strategies: List[ParseStrategy]) -> List[str]:
        statements = self._read_sql_statements(path)
        for statement in statements:
            for strategy in strategies:
                if strategy.match(statement):
                    operation = strategy.convert(statement, env)
                    if operation.executable():
                        yield operation.as_summary_str()



from typing import List, Tuple

import pandas as pd

from sqltest.datasource.reader import DatasetReader


class ExcelDatasetReader(DatasetReader):
    def __init__(self, data_path: str):
        self._data_file = pd.ExcelFile(data_path)
        self.source_sheets = [source for source in self._data_file.sheet_names if
                              source.startswith('source--')]
        self.target_sheets = [source for source in self._data_file.sheet_names if
                              source.startswith('target--')]

    def read_source_data(self) -> List[Tuple[str, pd.DataFrame]]:
        for sheet in self.source_sheets:
            yield from self._parse_table(sheet)

    def read_target_data(self) -> List[Tuple[str, pd.DataFrame]]:
        for sheet in self.target_sheets:
            yield from self._parse_table(sheet)

    def _parse_table(self, sheet):
        table_name = self._data_file.parse(sheet, header=None).iloc[0].values[0]
        data = self._data_file.parse(sheet, skiprows=1)
        yield table_name, data



import logging.config

logging_config = {

    'version': 1,
    'formatters': {
        'default': {
            'format': '%(asctime)s %(levelname)s: %(message)s',
            'datefmt': '%Y/%m/%d %H:%M:%S'
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'default',
            'level': 'INFO',
            'stream': 'ext://sys.stdout',
        }
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True
        },
        'glue': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False
        },
    }
}

logging.config.dictConfig(logging_config)



from setuptools import setup

if __name__ == '__main__':
    with open("README.md", "r") as fh:
        long_description = fh.read()

    setup(
        name="sqltest",
        install_requires=["openpyxl >= 3.0.9"]
    )



from typing import List, Tuple

import pandas as pd

from sqltest.datasource.reader import DatasetReader
from sqltest.parser.sql_reader import SqlFileReader
from sqltest.parser.strategy import ExecutionEnvironment, ExtractConstantStrategy, ExtractSQLStrategy, \
    VerifyVariableStrategy


class SqlEngine(object):
    def __init__(self, env_variables: dict):
        self._sql_parse_strategies = [ExtractSQLStrategy(), ExtractConstantStrategy(), VerifyVariableStrategy()]
        self._execution_env = ExecutionEnvironment(env_variables)

    def run(self, dataset_reader: DatasetReader, sql_file_path: str):
        self.source_dataset = dataset_reader.read_source_data()
        self.target_dataset = dataset_reader.read_target_data()
        statements = SqlFileReader().compile_sql_file(sql_file_path, self._execution_env, self._sql_parse_strategies)
        self._execute(statements)

    def _execute(self, statements: List[str]):
        raise NotImplementedError("This method is not implemented.")

    def get_target_tables(self) -> List[Tuple[str, pd.DataFrame]]:
        raise NotImplementedError("This method is not implemented.")

    def get_target_dataset(self) -> List[Tuple[str, pd.DataFrame]]:
        return self.target_dataset

    def verify_target_dataset(self):
        raise NotImplementedError("This method is not implemented.")



class Operation(object):
    def as_summary_str(self) -> str:
        raise NotImplementedError("this method is not implemented.")

    def executable(self) -> bool:
        raise NotImplementedError("this method is not implemented.")


class NothingOperation(Operation):
    def executable(self) -> bool:
        return False

    def as_summary_str(self) -> str:
        return ""


class VerifyVariableOperation(NothingOperation):
    def __init__(self, variable):
        self.variable = variable

    def as_summary_str(self) -> str:
        return f"--variable={self.variable}"


class RunnableOperation(Operation):
    def __init__(self, sql: str):
        self._sql = sql

    def as_summary_str(self) -> str:
        return self._sql

    def executable(self) -> bool:
        return True



from typing import List, Tuple

import pandas as pd


class DatasetReader(object):
    def read_source_data(self) -> List[Tuple[str, pd.DataFrame]]:
        raise NotImplementedError("This method is not implemented.")

    def read_target_data(self) -> List[Tuple[str, pd.DataFrame]]:
        raise NotImplementedError("This method is not implemented.")



from sqltest.datasource import ExcelDatasetReader


def excel_reader(data_path: str):
    def wrapper(func):
        def decorate(*args):
            reader = ExcelDatasetReader(data_path)
            func(*args, reader)

        return decorate

    return wrapper
